# -*- coding: utf-8 -*-


import numpy as np
import cv2

from threading import local
from wxbtool.data.variables import split_name
from wxbtool.util.cmaps import cmaps, var2cmap


data = local()


def imgdata():
    if "img" in dir(data):
        return data.img
    data.img = np.zeros([32, 64, 4], dtype=np.uint8)
    return data.img


def colorize(data, out, cmap):
    data = data.reshape(32, 64)
    data = (data - data.min() + 0.0001) / (data.max() - data.min() + 0.0001)
    data = (data * (data >= 0) * (data < 1) + (data >= 1)) * 255
    fliped = (data[::-1, :]).astype(np.uint8)
    return np.take(cmaps[cmap], fliped, axis=0, out=out)


def imsave(fileobj, data):
    is_success, img = cv2.imencode(".png", data)
    buffer = img.tobytes()
    fileobj.write(buffer)


def plot(var, fileobj, data):
    code, _ = split_name(var)
    imsave(fileobj, colorize(data, imgdata(), var2cmap[code]))



import numpy as np
import torch as th
import lightning as ltn

from torch.utils.data import DataLoader
from wxbtool.util.plotter import plot


class LightningModel(ltn.LightningModule):
    def __init__(self, model, opt=None):
        super(LightningModel, self).__init__()
        self.model = model
        self.learning_rate = 1e-3

        self.counter = 0
        self.labeled_loss = 0
        self.labeled_rmse = 0

        self.opt = opt

    def configure_optimizers(self):
        optimizer = th.optim.AdamW(self.parameters(), lr=self.learning_rate)
        scheduler = th.optim.lr_scheduler.CosineAnnealingLR(optimizer, 53)
        return [optimizer], [scheduler]

    def loss_fn(self, input, result, target):
        return self.model.lossfun(input, result, target)

    def compute_rmse(self, targets, results):
        _, tgt = self.model.get_targets(**targets)
        _, rst = self.model.get_results(**results)
        tgt = (
            tgt.detach().cpu().numpy().reshape(-1, self.model.setting.pred_span, 32, 64)
        )
        rst = (
            rst.detach().cpu().numpy().reshape(-1, self.model.setting.pred_span, 32, 64)
        )
        rmse = np.sqrt(np.mean(self.model.weight.cpu().numpy() * (rst - tgt) ** 2))
        return rmse

    def forward(self, **inputs):
        return self.model(**inputs)

    def plot(self, inputs, results, targets):
        vars_in, _ = self.model.get_inputs(**inputs)
        for bas, var in enumerate(self.model.setting.vars_in):
            for ix in range(self.model.setting.input_span):
                img = vars_in[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                plot(var, open("%s_inp_%d.png" % (var, ix), mode="wb"), img)

        vars_fc, _ = self.model.get_results(**results)
        vars_tg, _ = self.model.get_targets(**targets)
        for bas, var in enumerate(self.model.setting.vars_out):
            for ix in range(self.model.setting.pred_span):
                fcst = vars_fc[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                tgrt = vars_tg[var][0, ix].detach().cpu().numpy().reshape(32, 64)
                plot(var, open("%s_fcs_%d.png" % (var, ix), mode="wb"), fcst)
                plot(var, open("%s_tgt_%d.png" % (var, ix), mode="wb"), tgrt)

    def training_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)

        loss = self.loss_fn(inputs, results, targets)

        self.log("train_loss", loss, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)
        loss = self.loss_fn(inputs, results, targets)
        rmse = self.compute_rmse(targets, results)

        self.log("val_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("val_rmse", rmse, on_step=False, on_epoch=True, prog_bar=True)

        batch_len = inputs[self.model.setting.vars[0]].shape[0]
        self.labeled_loss += loss.item() * batch_len
        self.labeled_rmse += rmse * batch_len
        self.counter += batch_len

        self.plot(inputs, results, targets)

    def test_step(self, batch, batch_idx):
        inputs, targets = batch
        inputs = {v: inputs[v].float() for v in self.model.setting.vars}
        targets = {v: targets[v].float() for v in self.model.setting.vars}
        results = self.forward(**inputs)
        loss = self.loss_fn(inputs, results, targets)
        rmse = self.compute_rmse(targets, results)

        self.log("test_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("test_rmse", rmse, on_step=False, on_epoch=True, prog_bar=True)

        self.plot(inputs, results, targets)

    def on_save_checkpoint(self, checkpoint) -> None:
        import glob
        import os

        rmse = self.labeled_rmse / self.counter
        loss = self.labeled_loss / self.counter
        record = "%2.5f-%03d-%1.5f.ckpt" % (rmse, checkpoint["epoch"], loss)
        fname = "best-%s" % record
        with open(fname, "bw") as f:
            th.save(checkpoint, f)
        for ix, ckpt in enumerate(sorted(glob.glob("best-*.ckpt"), reverse=True)):
            if ix > 5:
                os.unlink(ckpt)

        self.counter = 0
        self.labeled_loss = 0
        self.labeled_correct = 0

        print()

    def train_dataloader(self):
        if self.model.dataset_train is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_train,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=True,
        )

    def val_dataloader(self):
        if self.model.dataset_eval is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_eval,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=False,
        )

    def test_dataloader(self):
        if self.model.dataset_test is None:
            if self.opt.data != "":
                self.model.load_dataset("train", "client", url=self.opt.data)
            else:
                self.model.load_dataset("train", "server")
        return DataLoader(
            self.model.dataset_test,
            batch_size=self.opt.batch_size,
            num_workers=self.opt.n_cpu,
            shuffle=False,
        )



# -*- coding: utf-8 -*-

"""
 A modeling spec for t850

 This spec follows basic settings and discussions in

   Improving Data‐Driven Global Weather Prediction Using Deep Convolutional Neural Networks on a Cubed Sphere
   by Jonathan A. Weyn, Dale R. Durran, Rich Caruana
   https://doi.org/10.1029/2020MS002109

"""

import torch as th
import torch.nn as nn
from wxbtool.nn.model import Base2d
from wxbtool.nn.setting import Setting
from wxbtool.norms.meanstd import (
    norm_z500,
    norm_z1000,
    norm_tau,
    norm_t850,
    norm_t2m,
    norm_tisr,
    denorm_t850,
)


mse = nn.MSELoss()


class SettingWeyn(Setting):
    def __init__(self):
        super().__init__()
        self.resolution = "5.625deg"  # The spatial resolution of the model

        self.levels = [
            "300",
            "500",
            "700",
            "850",
            "1000",
        ]  # Which vertical levels to choose
        self.height = len(self.levels)  # How many vertical levels to choose

        # The name of variables to choose, for both input features and output
        self.vars = [
            "geopotential",
            "toa_incident_solar_radiation",
            "2m_temperature",
            "temperature",
        ]

        # The code of variables in input features
        self.vars_in = ["z500", "z1000", "tau", "t850", "t2m", "tisr"]
        # The code of variables in output
        self.vars_out = ["t850"]

        # temporal scopes for train
        self.years_train = [
            1980,
            1981,
            1982,
            1983,
            1984,
            1985,
            1986,
            1987,
            1988,
            1989,
            1990,
            1991,
            1992,
            1993,
            1994,
            1995,
            1996,
            1997,
            1998,
            1999,
            2000,
            2001,
            2002,
            2003,
            2004,
            2005,
            2006,
            2007,
            2008,
            2009,
            2010,
            2011,
            2012,
            2013,
            2014,
        ]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting3d(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8  # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3  # How many hourly steps for an input
        self.pred_span = 1  # How many hourly steps for a prediction
        self.pred_shift = 72  # How many hours between the end of the input span and the beginning of prediction span


class SettingTest(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8  # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3  # How many hourly steps for an input
        self.pred_span = 1  # How many hourly steps for a prediction
        self.pred_shift = 72  # How many hours between the end of the input span and the beginning of prediction span

        # temporal scopes for train
        self.years_train = [2013, 2014]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting5d(SettingWeyn):
    def __init__(self):
        super().__init__()
        self.step = 8  # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3  # How many hourly steps for an input
        self.pred_span = 1  # How many hourly steps for a prediction
        self.pred_shift = 120  # How many hours between the end of the input span and the beginning of prediction span


class Spec(Base2d):
    def __init__(self, setting):
        super().__init__(setting)

        # following Weyn's schema from the above paper
        self.name = "t850_weyn"

    def get_inputs(self, **kwargs):
        z500 = norm_z500(
            kwargs["geopotential"].view(
                -1, self.setting.input_span, self.setting.height, 32, 64
            )[:, :, self.setting.levels.index("500")]
        )
        z1000 = norm_z1000(
            kwargs["geopotential"].view(
                -1, self.setting.input_span, self.setting.height, 32, 64
            )[:, :, self.setting.levels.index("1000")]
        )
        tau = norm_tau(
            kwargs["geopotential"].view(
                -1, self.setting.input_span, self.setting.height, 32, 64
            )[:, :, self.setting.levels.index("300")]
            - kwargs["geopotential"].view(
                -1, self.setting.input_span, self.setting.height, 32, 64
            )[:, :, self.setting.levels.index("700")]
        )
        t850 = norm_t850(
            kwargs["temperature"].view(
                -1, self.setting.input_span, self.setting.height, 32, 64
            )[:, :, self.setting.levels.index("850")]
        )
        t2m = norm_t2m(
            kwargs["2m_temperature"].view(-1, self.setting.input_span, 32, 64)
        )
        tisr = norm_tisr(
            kwargs["toa_incident_solar_radiation"].view(
                -1, self.setting.input_span, 32, 64
            )
        )

        z500 = self.augment_data(z500)
        z1000 = self.augment_data(z1000)
        tau = self.augment_data(tau)
        t850 = self.augment_data(t850)
        t2m = self.augment_data(t2m)
        tisr = self.augment_data(tisr)

        return {
            "z500": z500,
            "z1000": z1000,
            "tau": tau,
            "t850": t850,
            "t2m": t2m,
            "tisr": tisr,
        }, th.cat(
            (
                z500,
                z1000,
                tau,
                t850,
                t2m,
                tisr,
            ),
            dim=1,
        )

    def get_targets(self, **kwargs):
        t850 = kwargs["temperature"].view(-1, 1, self.setting.height, 32, 64)[
            :, :, self.setting.levels.index("850")
        ]
        t850 = self.augment_data(t850)
        return {"t850": t850}, t850

    def get_results(self, **kwargs):
        t850 = denorm_t850(kwargs["t850"])
        return {"t850": t850}, t850

    def forward(self, **kwargs):
        raise NotImplementedError("Spec is abstract and can not be initialized")

    def lossfun(self, inputs, result, target):
        _, rst = self.get_results(**result)
        _, tgt = self.get_targets(**target)
        rst = self.weight * rst.view(-1, 1, 32, 64)
        tgt = self.weight * tgt.view(-1, 1, 32, 64)

        losst = mse(rst[:, 0], tgt[:, 0])

        return losst



from arghandler import ArgumentHandler, subcmd

from wxbtool.data.dsserver import main as dsmain
from wxbtool.nn.train import main as tnmain
from wxbtool.nn.test import main as ttmain


@subcmd
def help(parser, context, args):
    pass


@subcmd("dserve", help="start the dataset server")
def dserve(parser, context, args):
    parser.add_argument(
        "-i", "--ip", type=str, default="127.0.0.1", help="the ip of the dataset serevr"
    )
    parser.add_argument(
        "-p", "--port", type=int, default=8088, help="the port of the dataset serevr"
    )
    parser.add_argument(
        "-w", "--workers", type=int, default=4, help="the number of workers"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.specs.res5_625.t850weyn",
        help="module of a metrological model to load",
    )
    parser.add_argument(
        "-s",
        "--setting",
        type=str,
        default="Setting",
        help="setting for a metrological model spec",
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )
    opt = parser.parse_args(args)

    dsmain(context, opt)


@subcmd("train", help="start training")
def train(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default="0", help="index of gpu")
    parser.add_argument(
        "-c",
        "--n_cpu",
        type=int,
        default=8,
        help="number of cpu threads to use during batch generation",
    )
    parser.add_argument(
        "-b", "--batch_size", type=int, default=64, help="size of the batches"
    )
    parser.add_argument(
        "-e",
        "--epoch",
        type=int,
        default=0,
        help="current epoch to start training from",
    )
    parser.add_argument(
        "-n", "--n_epochs", type=int, default=200, help="number of epochs of training"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.zoo.unet.t850d3",
        help="module of the metrological model to load",
    )
    parser.add_argument(
        "-l",
        "--load",
        type=str,
        default="",
        help="dump file of the metrological model to load",
    )
    parser.add_argument(
        "-k", "--check", type=str, default="", help="checkpoint file to load"
    )
    parser.add_argument("-r", "--rate", type=float, default=0.001, help="learning rate")
    parser.add_argument(
        "-w", "--weightdecay", type=float, default=0.0, help="weight decay"
    )
    parser.add_argument(
        "-d", "--data", type=str, default="", help="url of the dataset server"
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )
    opt = parser.parse_args(args)

    tnmain(context, opt)


@subcmd("test", help="start testing")
def test(parser, context, args):
    parser.add_argument("-g", "--gpu", type=str, default="0", help="index of gpu")
    parser.add_argument(
        "-c",
        "--n_cpu",
        type=int,
        default=8,
        help="number of cpu threads to use during batch generation",
    )
    parser.add_argument(
        "-b", "--batch_size", type=int, default=64, help="size of the batches"
    )
    parser.add_argument(
        "-m",
        "--module",
        type=str,
        default="wxbtool.zoo.unet.t850d3",
        help="module of the metrological model to load",
    )
    parser.add_argument(
        "-l",
        "--load",
        type=str,
        default="",
        help="dump file of the metrological model to load",
    )
    parser.add_argument(
        "-d", "--data", type=str, default="", help="url of the dataset server"
    )
    parser.add_argument(
        "-t", "--test", type=str, default="false", help="setting for test"
    )
    opt = parser.parse_args(args)

    ttmain(context, opt)


def main():
    import sys

    handler = ArgumentHandler()
    if len(sys.argv) < 2:
        handler.run(["help"])
    else:
        handler.run(sys.argv[1:])



# -*- coding: utf-8 -*-

import os
import os.path as path
import hashlib
import requests
import logging
import json

import xarray as xr
import numpy as np

import msgpack
import msgpack_numpy as m

m.patch()

from itertools import product  # noqa: E402
from torch.utils.data import Dataset  # noqa: E402

from wxbtool.data.variables import codes, vars2d, vars3d  # noqa: E402

logger = logging.getLogger()

all_levels = [
    "50",
    "100",
    "150",
    "200",
    "250",
    "300",
    "400",
    "500",
    "600",
    "700",
    "850",
    "925",
    "1000",
]


class WindowArray(type(np.zeros(0, dtype=np.float32))):

    def __new__(subtype, orig, shift=0, step=1):
        shape = [orig.shape[_] for _ in range(len(orig.shape))]
        self = np.ndarray.__new__(
            subtype, shape, dtype=np.float32, buffer=orig.tobytes()
        )[shift::step]
        return self


class WxDataset(Dataset):
    def __init__(
        self,
        root,
        resolution,
        years,
        vars,
        levels,
        step=1,
        input_span=2,
        pred_shift=24,
        pred_span=1,
    ):
        self.root = root
        self.resolution = resolution
        self.input_span = input_span
        self.step = step
        self.pred_shift = pred_shift
        self.pred_span = pred_span
        self.years = years
        self.vars = vars
        self.levels = levels
        self.inputs = {}
        self.targets = {}
        self.shapes = {
            "data": {},
        }
        self.accumulated = {}

        if resolution == "5.625deg":
            self.height = 13
            self.width = 32
            self.length = 64

        code = "%s:%s:%s:%s:%s:%s:%s:%s" % (
            resolution,
            years,
            vars,
            levels,
            step,
            input_span,
            pred_shift,
            pred_span,
        )
        hashstr = hashlib.md5(code.encode("utf-8")).hexdigest()
        self.hashcode = hashstr

        dumpdir = path.abspath("%s/.cache/%s" % (self.root, hashstr))
        if not path.exists(dumpdir):
            os.makedirs(dumpdir)
            self.load(dumpdir)

        self.memmap(dumpdir)

    def load(self, dumpdir):
        levels_selector = [all_levels.index(lvl) for lvl in self.levels]
        selector = np.array(levels_selector, dtype=np.int64)

        size = 0
        lastvar = None
        for var, yr in product(self.vars, self.years):
            if var in vars3d:
                length = self.load_3ddata(yr, var, selector, self.accumulated)
            elif var in vars2d:
                length = self.load_2ddata(yr, var, self.accumulated)
            else:
                raise ValueError("variable %s does not supported!" % var)
            size += length

            if lastvar and lastvar != var:
                self.inputs[lastvar] = WindowArray(
                    self.accumulated[lastvar],
                    shift=self.input_span * self.step,
                    step=self.step,
                )
                self.targets[lastvar] = WindowArray(
                    self.accumulated[lastvar],
                    shift=self.pred_span * self.step + self.pred_shift,
                    step=self.step,
                )
                self.dump_var(dumpdir, lastvar)

            lastvar = var

        if lastvar:
            self.inputs[lastvar] = WindowArray(
                self.accumulated[lastvar],
                shift=self.input_span * self.step,
                step=self.step,
            )
            self.targets[lastvar] = WindowArray(
                self.accumulated[lastvar],
                shift=self.pred_span * self.step + self.pred_shift,
                step=self.step,
            )
            self.dump_var(dumpdir, lastvar)

        if not self.accumulated:
            logger.error("No data accumulated. Please check the data loading process.")
            raise ValueError(
                "No data accumulated. Ensure that data is correctly loaded and accumulated."
            )

        lengths = {var: acc.shape[0] for var, acc in self.accumulated.items()}
        unique_lengths = set(lengths.values())

        if len(unique_lengths) != 1:
            max_length = max(unique_lengths)

            inconsistent_vars = {
                var: length for var, length in lengths.items() if length != max_length
            }

            if inconsistent_vars:
                for var, length in inconsistent_vars.items():
                    logger.error(
                        f"Variable {var} has inconsistent length {length}. Expected length: {max_length}."
                    )

                # 根据需要，选择是否抛出异常
                raise ValueError(
                    "Inconsistent data lengths across variables detected. Please check the data loading process."
                )
        else:
            logger.info("All variables have consistent data lengths.")

        with open("%s/shapes.json" % dumpdir, mode="w") as fp:
            json.dump(self.shapes, fp)

        self.size = size // len(self.vars)
        logger.info("total %s items loaded!", self.size)

        for var in list(self.accumulated.keys()):
            del self.accumulated[var]

    def load_2ddata(self, year, var, accumulated):
        data_path = "%s/%s/%s_%d_%s.nc" % (self.root, var, var, year, self.resolution)
        with xr.open_dataset(data_path) as ds:
            ds = ds.transpose("time", "lat", "lon")
            if var not in accumulated:
                accumulated[var] = np.array(ds[codes[var]].data, dtype=np.float32)
            else:
                accumulated[var] = np.concatenate(
                    [accumulated[var], np.array(ds[codes[var]].data, dtype=np.float32)],
                    axis=0,
                )
            logger.info("%s[%d]: %s", var, year, str(accumulated[var].shape))

        return accumulated[var].shape[0]

    def load_3ddata(self, year, var, selector, accumulated):
        data_path = "%s/%s/%s_%d_%s.nc" % (self.root, var, var, year, self.resolution)
        with xr.open_dataset(data_path) as ds:
            ds = ds.transpose("time", "level", "lat", "lon")
            if var not in accumulated:
                accumulated[var] = np.array(ds[codes[var]].data, dtype=np.float32)[
                    :, selector, :, :
                ]
            else:
                accumulated[var] = np.concatenate(
                    [
                        accumulated[var],
                        np.array(ds[codes[var]].data, dtype=np.float32)[
                            :, selector, :, :
                        ],
                    ],
                    axis=0,
                )
            logger.info("%s[%d]: %s", var, year, str(accumulated[var].shape))

        return accumulated[var].shape[0]

    def dump_var(self, dumpdir, var):
        file_dump = "%s/%s.npy" % (dumpdir, var)
        self.shapes["data"][var] = self.accumulated[var].shape
        np.save(file_dump, self.accumulated[var])

    def memmap(self, dumpdir):
        with open("%s/shapes.json" % dumpdir) as fp:
            shapes = json.load(fp)

        for var in self.vars:
            file_dump = "%s/%s.npy" % (dumpdir, var)

            # load data from memmap, and skip the first shift elements of mmap data header
            shape = shapes["data"][var]
            total_size = np.prod(shape)
            data = np.memmap(file_dump, dtype=np.float32, mode="r")
            shift = data.shape[0] - total_size
            self.accumulated[var] = np.reshape(data[shift:], shape)

            if var in vars2d or var in vars3d:
                self.inputs[var] = self.accumulated[var]
                self.targets[var] = self.accumulated[var]

    def __len__(self):
        length = (
            self.accumulated[self.vars[0]].shape[0]
            - (self.input_span - 1) * self.step
            - (self.pred_span - 1) * self.step
            - self.pred_shift
        )
        logger.info(f"Dataset length: {length}")
        return length

    def __getitem__(self, item):
        inputs, targets = {}, {}
        for var in self.vars:
            if var in vars2d or var in vars3d:
                input_slice = self.inputs[var][item :: self.step][: self.input_span]
                target_slice = self.targets[var][
                    item
                    + self.step * (self.input_span - 1)
                    + self.pred_shift :: self.step
                ][: self.pred_span]
                inputs[var] = input_slice
                targets[var] = target_slice
                if input_slice.shape[0] != self.input_span:
                    logger.warning(
                        f"Input slice for var {var} at index {item} has shape {input_slice.shape}"
                    )
                if target_slice.shape[0] != self.pred_span:
                    logger.warning(
                        f"Target slice for var {var} at index {item} has shape {target_slice.shape}"
                    )

        return inputs, targets


class WxDatasetClient(Dataset):
    def __init__(
        self,
        url,
        phase,
        resolution,
        years,
        vars,
        levels,
        step=1,
        input_span=2,
        pred_shift=24,
        pred_span=1,
    ):
        self.url = url
        self.phase = phase

        code = "%s:%s:%s:%s:%s:%s:%s:%s" % (
            resolution,
            years,
            vars,
            levels,
            step,
            input_span,
            pred_shift,
            pred_span,
        )
        self.hashcode = hashlib.md5(code.encode("utf-8")).hexdigest()

    def __len__(self):
        url = "%s/%s/%s" % (self.url, self.hashcode, self.phase)
        r = requests.get(url)
        if r.status_code != 200:
            raise Exception("http error %s: %s" % (r.status_code, r.text))

        data = msgpack.loads(r.content)

        return data["size"]

    def __getitem__(self, item):
        url = "%s/%s/%s/%d" % (self.url, self.hashcode, self.phase, item)
        r = requests.get(url)
        if r.status_code != 200:
            raise Exception("http error %s: %s" % (r.status_code, r.text))

        data = msgpack.loads(r.content)
        for key, val in data.items():
            for var, blk in val.items():
                val[var] = np.array(np.copy(blk), dtype=np.float32)

        return data["inputs"], data["targets"]



# -*- coding: utf-8 -*-

"""
 A modeling spec for t850

 This spec follows basic settings and discussions in

   Data-driven medium-range weather prediction with a Resnet pretrained on climate simulations:
   A new model for WeatherBench
   by Stephan Rasp, Nils Thuerey
   https://arxiv.org/pdf/2008.08626.pdf

"""

import torch as th
import torch.nn as nn
from wxbtool.nn.model import Base2d
from wxbtool.nn.setting import Setting
from wxbtool.data.variables import vars3d, code2var, split_name
from wxbtool.norms.meanstd import normalizors, denorm_t850


mse = nn.MSELoss()


class SettingRasp(Setting):
    def __init__(self):
        super().__init__()
        self.resolution = "5.625deg"  # The spatial resolution of the model

        # Which vertical levels to choose
        self.levels = ["50", "250", "500", "600", "700", "850", "925"]
        # How many vertical levels to choose
        self.height = len(self.levels)

        # The name of variables to choose, for both input features and output
        self.vars = [
            "geopotential",
            "temperature",
            "specific_humidity",
            "u_component_of_wind",
            "v_component_of_wind",
            "2m_temperature",
            "total_precipitation",
            "toa_incident_solar_radiation",
        ]

        # The code of variables in input features
        self.vars_in = [
            "z50",
            "z250",
            "z500",
            "z600",
            "z700",
            "z850",
            "z925",
            "t50",
            "t250",
            "t500",
            "t600",
            "t700",
            "t850",
            "t925",
            "q50",
            "q250",
            "q500",
            "q600",
            "q700",
            "q850",
            "q925",
            "u50",
            "u250",
            "u500",
            "u600",
            "u700",
            "u850",
            "u925",
            "v50",
            "v250",
            "v500",
            "v600",
            "v700",
            "v850",
            "v925",
            "t2m",
            "tp",
            "tisr",
        ]

        # The code of variables in output
        self.vars_out = ["t850"]

        # temporal scopes for train
        self.years_train = [
            1980,
            1981,
            1982,
            1983,
            1984,
            1985,
            1986,
            1987,
            1988,
            1989,
            1990,
            1991,
            1992,
            1993,
            1994,
            1995,
            1996,
            1997,
            1998,
            1999,
            2000,
            2001,
            2002,
            2003,
            2004,
            2005,
            2006,
            2007,
            2008,
            2009,
            2010,
            2011,
            2012,
            2013,
            2014,
        ]
        # temporal scopes for evaluation
        self.years_eval = [2015, 2016]
        # temporal scopes for test
        self.years_test = [2017, 2018]


class Setting3d(SettingRasp):
    def __init__(self):
        super().__init__()
        self.step = 8  # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3  # How many hourly steps for an input
        self.pred_span = 1  # How many hourly steps for a prediction
        self.pred_shift = 72  # How many hours between the end of the input span and the beginning of prediction span


class Setting5d(SettingRasp):
    def __init__(self):
        super().__init__()
        self.step = 8  # How many hours of a hourly step which all features in organized temporally
        self.input_span = 3  # How many hourly steps for an input
        self.pred_span = 1  # How many hourly steps for a prediction
        self.pred_shift = 120  # How many hours between the end of the input span and the beginning of prediction span


class Spec(Base2d):
    def __init__(self, setting):
        super().__init__(setting)

        # following Rasp's schema from the above paper
        self.name = "t850_rasp"

    def get_inputs(self, **kwargs):
        vdic, vlst = {}, []
        for nm in self.setting.vars_in:
            code, lvl = split_name(nm)
            v = code2var[code]
            if v in vars3d:
                d = kwargs[v].view(
                    -1, self.setting.input_span, self.setting.height, 32, 64
                )[:, :, self.setting.levels.index(lvl)]
            else:
                d = kwargs[v].view(-1, self.setting.input_span, 32, 64)
            d = normalizors[nm](d)
            d = self.augment_data(d)
            vdic[nm] = d
            vlst.append(d)

        return vdic, th.cat(vlst, dim=1)

    def get_targets(self, **kwargs):
        t850 = kwargs["temperature"].view(-1, 1, self.setting.height, 32, 64)[
            :, :, self.setting.levels.index("850")
        ]
        t850 = self.augment_data(t850)
        return {"t850": t850}, t850

    def get_results(self, **kwargs):
        t850 = denorm_t850(kwargs["t850"])
        return {"t850": t850}, t850

    def forward(self, **kwargs):
        raise NotImplementedError("Spec is abstract and can not be initialized")

    def lossfun(self, inputs, result, target):
        _, rst = self.get_results(**result)
        _, tgt = self.get_targets(**target)
        rst = self.weight * rst.view(-1, 1, 32, 64)
        tgt = self.weight * tgt.view(-1, 1, 32, 64)

        losst = mse(rst[:, 0], tgt[:, 0])

        return losst
